# -*- coding: utf-8 -*-
"""

"""
from numpy import array, zeros, dot, ndarray, ix_, outer

from pyfem.elements.BaseElement import BaseElement
from pyfem.elements.set_element_field_variables import set_element_field_variables
from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Dof import Dof
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.isoelements.IsoElementShape import IsoElementShape
from pyfem.materials.get_material_data import MaterialData
from pyfem.utils.colors import error_style


class SolidPhaseDamageSmallStrain(BaseElement):
    r"""
    固体相场断裂单元。

    :ivar qp_b_matrices: 积分点处的B矩阵列表
    :vartype qp_b_matrices: ndarray

    :ivar qp_b_matrices_transpose: 积分点处的B矩阵转置列表
    :vartype qp_b_matrices_transpose: ndarray

    :ivar qp_strains: 积分点处的应变列表
    :vartype qp_strains: list[ndarray]

    :ivar qp_stresses: 积分点处的应力列表
    :vartype qp_stresses: list[ndarray]

    :ivar qp_phases: 积分点处的相场变量列表
    :vartype qp_phases: list[ndarray]

    :ivar qp_phase_fluxes: 积分点处的相场变量通量列表
    :vartype qp_phase_fluxes: list[ndarray]

    :ivar qp_ddsddps: 积分点处的相场刚度矩阵列表
    :vartype qp_ddsddps: list[ndarray]

    :ivar dof_u: 单元位移自由度列表
    :vartype dof_u: list[int]

    :ivar dof_p: 单元相场自由度列表
    :vartype dof_p: list[int]

    :ivar ntens: 总应力数量
    :vartype ntens: int

    :ivar ndi: 轴向应力数量
    :vartype ndi: int

    :ivar nshr: 剪切应力数量
    :vartype nshr: int

    .. math::
        K_{ij}^\varphi  = \int_V {\left[ {{g_c}{l_c}{{\left( {\nabla {N_i}} \right)}^T}\nabla {N_j} + \left( {\frac{{{g_c}}}{{{l_c}}} + 2H} \right){N_i}{N_j}} \right]{\text{d}}V}

    .. math::
        RHS_i^\varphi  = \int_V {\left[ {{g_c}{l_c}{{\left( {\nabla {N_i}} \right)}^T}\nabla \varphi  - \left( {2\left( {1 - \varphi } \right)H - \frac{{{g_c}}}{{{l_c}}}\varphi } \right){N_i}} \right]{\text{d}}V}
    """

    __slots_dict__: dict = {
        'qp_b_matrices': ('ndarray', '积分点处的B矩阵列表'),
        'qp_b_matrices_transpose': ('ndarray', '积分点处的B矩阵转置列表'),
        'qp_strains': ('list[ndarray]', '积分点处的应变列表'),
        'qp_stresses': ('list[ndarray]', '积分点处的应力列表'),
        'qp_phases': ('list[ndarray]', '积分点处的相场变量列表'),
        'qp_phase_fluxes': ('list[ndarray]', '积分点处的相场变量通量列表'),
        'qp_ddsddps': ('list[ndarray]', '积分点处的相场刚度矩阵列表'),
        'qp_energies': ('list[ndarray]', '积分点处的相场刚度矩阵列表'),
        'dof_u': ('list[int]', '单元位移自由度列表'),
        'dof_p': ('list[int]', '单元相场自由度列表'),
        'ntens': ('int', '总应力数量'),
        'ndi': ('int', '轴向应力数量'),
        'nshr': ('int', '剪切应力数量')
    }

    __slots__ = BaseElement.__slots__ + [slot for slot in __slots_dict__.keys()]

    __allowed_material_data_list__ = [('ElasticIsotropic', 'PlasticKinematicHardening', 'ViscoElasticMaxwell', 'PlasticCrystal', 'PlasticCrystalGNDs', 'User'),
                                      ('PhaseFieldDamage', 'User')]

    def __init__(self, element_id: int,
                 iso_element_shape: IsoElementShape,
                 connectivity: ndarray,
                 node_coords: ndarray,
                 dof: Dof,
                 materials: list[Material],
                 section: Section,
                 material_data_list: list[MaterialData],
                 timer: Timer) -> None:

        super().__init__(element_id, iso_element_shape, connectivity, node_coords)

        self.allowed_material_data_list = self.__allowed_material_data_list__
        self.allowed_material_number = len(self.allowed_material_data_list)

        self.dof = dof
        self.materials = materials
        self.section = section
        self.material_data_list = material_data_list
        self.check_materials()
        self.timer = timer

        if self.dimension == 2:
            self.dof_names = ['u1', 'u2', 'phi']
            self.ntens = 4
            self.ndi = 3
            self.nshr = 1
        elif self.dimension == 3:
            self.dof_names = ['u1', 'u2', 'u3', 'phi']
            self.ntens = 6
            self.ndi = 3
            self.nshr = 3
        else:
            error_msg = f'{self.dimension} is not the supported dimension'
            raise NotImplementedError(error_style(error_msg))

        if dof.names != self.dof_names:
            error_msg = f'{dof.names} is not the supported dof of {type(self).__name__} element'
            raise NotImplementedError(error_style(error_msg))

        element_dof_number = len(self.dof_names) * self.iso_element_shape.nodes_number
        self.element_dof_number = element_dof_number
        self.element_dof_values = zeros(element_dof_number, dtype=DTYPE)
        self.element_ddof_values = zeros(element_dof_number, dtype=DTYPE)
        self.element_fint = zeros(element_dof_number, dtype=DTYPE)
        self.element_stiffness = zeros(shape=(self.element_dof_number, self.element_dof_number), dtype=DTYPE)

        self.qp_b_matrices: ndarray = None  # type: ignore
        self.qp_b_matrices_transpose: ndarray = None  # type: ignore
        self.qp_strains: list[ndarray] = None  # type: ignore
        self.qp_stresses: list[ndarray] = None  # type: ignore
        self.qp_phases: list[ndarray] = None  # type: ignore
        self.qp_phase_fluxes: list[ndarray] = None  # type: ignore
        self.qp_ddsddps: list[ndarray] = None  # type: ignore
        self.qp_energies: list[ndarray] = None  # type: ignore

        for i in range(self.qp_number):
            self.qp_state_variables[i]['history_energy'] = array([0.0])
            self.qp_state_variables_new[i]['history_energy'] = array([0.0])

        self.dof_u: list[int] = list()
        self.dof_p: list[int] = list()
        for i in range(self.iso_element_shape.nodes_number):
            if self.dimension == 2:
                self.dof_u += [len(self.dof_names) * i + 0, len(self.dof_names) * i + 1]
                self.dof_p += [len(self.dof_names) * i + 2]
            elif self.dimension == 3:
                self.dof_u += [len(self.dof_names) * i + 0, len(self.dof_names) * i + 1, len(self.dof_names) * i + 2]
                self.dof_p += [len(self.dof_names) * i + 3]
        self.create_qp_b_matrices()

    def create_qp_b_matrices(self) -> None:
        if self.dimension == 2:
            self.qp_b_matrices = zeros(shape=(self.qp_number, 3, len(self.dof_u)), dtype=DTYPE)
            for iqp, qp_dhdx in enumerate(self.qp_dhdxes):
                for i, val in enumerate(qp_dhdx.transpose()):
                    self.qp_b_matrices[iqp, 0, i * 2 + 0] = val[0]
                    self.qp_b_matrices[iqp, 1, i * 2 + 1] = val[1]
                    self.qp_b_matrices[iqp, 2, i * 2 + 0] = val[1]
                    self.qp_b_matrices[iqp, 2, i * 2 + 1] = val[0]

        elif self.dimension == 3:
            self.qp_b_matrices = zeros(shape=(self.iso_element_shape.qp_number, 6, len(self.dof_u)), dtype=DTYPE)
            for iqp, qp_dhdx in enumerate(self.qp_dhdxes):
                for i, val in enumerate(qp_dhdx.transpose()):
                    self.qp_b_matrices[iqp, 0, i * 3 + 0] = val[0]
                    self.qp_b_matrices[iqp, 1, i * 3 + 1] = val[1]
                    self.qp_b_matrices[iqp, 2, i * 3 + 2] = val[2]
                    self.qp_b_matrices[iqp, 3, i * 3 + 0] = val[1]
                    self.qp_b_matrices[iqp, 3, i * 3 + 1] = val[0]
                    self.qp_b_matrices[iqp, 4, i * 3 + 0] = val[2]
                    self.qp_b_matrices[iqp, 4, i * 3 + 2] = val[0]
                    self.qp_b_matrices[iqp, 5, i * 3 + 1] = val[2]
                    self.qp_b_matrices[iqp, 5, i * 3 + 2] = val[1]

        self.qp_b_matrices_transpose = array([qp_b_matrix.transpose() for qp_b_matrix in self.qp_b_matrices])

    def update_element_material_stiffness_fint(self,
                                               is_update_material: bool = True,
                                               is_update_stiffness: bool = True,
                                               is_update_fint: bool = True, ) -> None:
        element_id = self.element_id
        timer = self.timer
        ntens = self.ntens
        ndi = self.ndi
        nshr = self.nshr

        dimension = self.iso_element_shape.dimension

        qp_number = self.qp_number
        qp_shape_values = self.iso_element_shape.qp_shape_values
        qp_shape_gradients = self.iso_element_shape.qp_shape_gradients
        qp_dhdxes = self.qp_dhdxes

        qp_b_matrices = self.qp_b_matrices
        qp_b_matrices_transpose = self.qp_b_matrices_transpose
        qp_jacobi_invs = self.qp_jacobi_invs
        qp_weight_times_jacobi_dets = self.qp_weight_times_jacobi_dets

        qp_state_variables = self.qp_state_variables
        qp_state_variables_new = self.qp_state_variables_new

        element_dof_values = self.element_dof_values
        element_ddof_values = self.element_ddof_values

        u = element_dof_values[self.dof_u]
        phi = element_dof_values[self.dof_p]

        du = element_ddof_values[self.dof_u]
        dphi = element_ddof_values[self.dof_p]

        solid_material_data = self.material_data_list[0]
        phase_material_data = self.material_data_list[1]

        gc = phase_material_data.gc  # type: ignore
        lc = phase_material_data.lc  # type: ignore

        if is_update_stiffness:
            self.element_stiffness = zeros(shape=(self.element_dof_number, self.element_dof_number), dtype=DTYPE)

        if is_update_fint:
            self.element_fint = zeros(self.element_dof_number, dtype=DTYPE)

        if is_update_material:
            self.qp_ddsddes = list()
            self.qp_strains = list()
            self.qp_stresses = list()
            self.qp_ddsddps = list()
            self.qp_phases = list()
            self.qp_phase_fluxes = list()
            self.qp_energies = list()

        for i in range(qp_number):
            if is_update_material:
                qp_weight_times_jacobi_det = qp_weight_times_jacobi_dets[i]
                qp_shape_value = qp_shape_values[i]
                qp_dhdx = qp_dhdxes[i]
                qp_b_matrix_transpose = qp_b_matrices_transpose[i]
                qp_b_matrix = qp_b_matrices[i]
                qp_strain = dot(qp_b_matrix, u)
                qp_dstrain = dot(qp_b_matrix, du)
                qp_phase = dot(qp_shape_value, phi)
                qp_dphase = dot(qp_shape_value, dphi)
                qp_phase_gradient = dot(qp_dhdx, phi)
                qp_dphase_gradient = dot(qp_dhdx, dphi)

                qp_degradation = (1.0 - qp_phase) ** 2 + 1.0e-8
                qp_degradation = min(qp_degradation, 1.0)
                qp_degradation = max(qp_degradation, 0.0)

                variable = {'strain': qp_strain, 'dstrain': qp_dstrain}
                qp_ddsdde, qp_output = solid_material_data.get_tangent(variable=variable,
                                                                       state_variable=qp_state_variables[i],
                                                                       state_variable_new=qp_state_variables_new[i],
                                                                       element_id=element_id,
                                                                       iqp=i,
                                                                       ntens=ntens,
                                                                       ndi=ndi,
                                                                       nshr=nshr,
                                                                       timer=timer)
                qp_stress = qp_output['stress']
                qp_strain_energy = qp_output['strain_energy']
                self.qp_ddsddes.append(qp_ddsdde)
                self.qp_strains.append(qp_strain)
                self.qp_stresses.append(qp_stress * qp_degradation)
                self.qp_phases.append(qp_phase)

            else:
                qp_b_matrix_transpose = qp_b_matrices_transpose[i]
                qp_b_matrix = qp_b_matrices[i]
                qp_weight_times_jacobi_det = qp_weight_times_jacobi_dets[i]
                qp_shape_value = qp_shape_values[i]
                qp_dhdx = qp_dhdxes[i]
                qp_ddsdde = self.qp_ddsddes[i]
                qp_stress = self.qp_stresses[i]
                qp_strain = dot(qp_b_matrix, u)
                qp_dstrain = dot(qp_b_matrix, du)
                qp_phase = dot(qp_shape_value, phi)
                qp_dphase = dot(qp_shape_value, dphi)
                qp_phase_gradient = dot(qp_dhdx, phi)
                qp_dphase_gradient = dot(qp_dhdx, dphi)

                qp_degradation = (1.0 - qp_phase) ** 2 + 1.0e-8
                qp_degradation = min(qp_degradation, 1.0)
                qp_degradation = max(qp_degradation, 0.0)

            # energy_positive, energy_negative = get_decompose_energy(qp_strain + qp_dstrain, qp_stress, dimension)
            # energy_positive += qp_strain_energy

            energy_positive = 0.5 * sum((qp_strain + qp_dstrain) * qp_stress)

            if energy_positive < qp_state_variables[i]['history_energy'][0]:
                energy_positive = qp_state_variables[i]['history_energy'][0]

            if energy_positive < qp_state_variables_new[i]['history_energy'][0]:
                energy_positive = qp_state_variables_new[i]['history_energy'][0]

            qp_state_variables_new[i]['history_energy'][0] = energy_positive

            self.qp_energies.append(energy_positive)

            if is_update_stiffness:
                self.element_stiffness[ix_(self.dof_u, self.dof_u)] += qp_weight_times_jacobi_det * \
                                                                       dot(qp_b_matrix_transpose, dot(qp_ddsdde * qp_degradation, qp_b_matrix))

                self.element_stiffness[ix_(self.dof_p, self.dof_p)] += qp_weight_times_jacobi_det * \
                                                                       ((gc / lc + 2.0 * energy_positive) * outer(qp_shape_value, qp_shape_value) +
                                                                        gc * lc * dot(qp_dhdx.transpose(), qp_dhdx))

                # vecu = -2.0 * (1.0 - (qp_phase + qp_dphase)) * dot(qp_b_matrix_transpose, qp_stress * qp_degradation) * qp_weight_times_jacobi_det
                # self.element_stiffness[ix_(self.dof_u, self.dof_p)] += outer(vecu, qp_shape_value)
                # self.element_stiffness[ix_(self.dof_p, self.dof_u)] += outer(qp_shape_value, vecu)

            if is_update_fint:
                self.element_fint[self.dof_u] += dot(qp_b_matrix_transpose, qp_stress * qp_degradation) * qp_weight_times_jacobi_det

                self.element_fint[self.dof_p] += qp_weight_times_jacobi_det * \
                                                 (gc * lc * dot(qp_dhdx.transpose(), (qp_phase_gradient + qp_dphase_gradient)) +
                                                  gc / lc * (qp_phase + qp_dphase) * qp_shape_value +
                                                  2.0 * ((qp_phase + qp_dphase) - 1.0) * energy_positive * qp_shape_value)

    def update_element_field_variables(self) -> None:
        self.qp_field_variables['strain'] = array(self.qp_strains, dtype=DTYPE) + array(self.qp_dstrains, dtype=DTYPE)
        self.qp_field_variables['stress'] = array(self.qp_stresses, dtype=DTYPE)
        self.qp_field_variables['energy'] = array(self.qp_energies, dtype=DTYPE)
        for key in self.qp_state_variables_new[0].keys():
            if key not in ['strain', 'stress', 'energy']:
                variable = []
                for qp_state_variable_new in self.qp_state_variables_new:
                    variable.append(qp_state_variable_new[key])
                self.qp_field_variables[f'SDV-{key}'] = array(variable, dtype=DTYPE)
        self.element_nodal_field_variables = set_element_field_variables(self.qp_field_variables, self.iso_element_shape, self.dimension)


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(SolidPhaseDamageSmallStrain.__slots_dict__)

    from pyfem.Job import Job

    job = Job(r'..\..\..\examples\mechanical_phase\rectangle\Job-1.toml')

    job.assembly.element_data_list[0].show()



# -*- coding: utf-8 -*-
"""

"""
from numpy import array, zeros, dot, ndarray

from pyfem.elements.BaseElement import BaseElement
from pyfem.elements.set_element_field_variables import set_element_field_variables
from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Dof import Dof
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.isoelements.IsoElementShape import IsoElementShape
from pyfem.materials.get_material_data import MaterialData
from pyfem.utils.colors import error_style


class Thermal(BaseElement):
    """
    温度单元。

    :ivar qp_temperatures: 积分点处的温度列表
    :vartype qp_temperatures: ndarray

    :ivar qp_heat_fluxes: 积分点处的热流密度列表
    :vartype qp_heat_fluxes: ndarray

    :ivar qp_ddsddts: 积分点处的材料热传导系数矩阵列表
    :vartype qp_ddsddts: list[ndarray]

    :ivar ntens: 总应力数量
    :vartype ntens: int

    :ivar ndi: 轴向应力数量
    :vartype ndi: int

    :ivar nshr: 剪切应力数量
    :vartype nshr: int
    """

    __slots_dict__: dict = {
        'qp_temperatures': ('ndarray', '积分点处的温度列表'),
        'qp_heat_fluxes': ('ndarray', '积分点处的热流密度列表'),
        'qp_ddsddts': ('list[ndarray]', '积分点处的材料热传导系数矩阵列表'),
        'ntens': ('int', '总应力数量'),
        'ndi': ('int', '轴向应力数量'),
        'nshr': ('int', '剪切应力数量')
    }

    __slots__: list = BaseElement.__slots__ + [slot for slot in __slots_dict__.keys()]

    __allowed_material_data_list__ = [('ThermalIsotropic', 'User')]

    def __init__(self, element_id: int,
                 iso_element_shape: IsoElementShape,
                 connectivity: ndarray,
                 node_coords: ndarray,
                 dof: Dof,
                 materials: list[Material],
                 section: Section,
                 material_data_list: list[MaterialData],
                 timer: Timer) -> None:

        super().__init__(element_id, iso_element_shape, connectivity, node_coords)

        self.allowed_material_data_list = self.__allowed_material_data_list__
        self.allowed_material_number = 1

        self.dof = dof
        self.materials = materials
        self.section = section
        self.material_data_list = material_data_list
        self.check_materials()
        self.timer = timer

        self.dof_names = ['T']
        self.ntens = 6
        self.ndi = 3
        self.nshr = 3

        if dof.names != self.dof_names:
            error_msg = f'{dof.names} is not the supported dof of {type(self).__name__} element'
            raise NotImplementedError(error_style(error_msg))

        element_dof_number = len(self.dof_names) * self.iso_element_shape.nodes_number

        self.element_dof_number = element_dof_number
        self.element_dof_values = zeros(element_dof_number, dtype=DTYPE)
        self.element_ddof_values = zeros(element_dof_number, dtype=DTYPE)
        self.element_fint = zeros(element_dof_number, dtype=DTYPE)
        self.element_stiffness = zeros(shape=(self.element_dof_number, self.element_dof_number), dtype=DTYPE)

        self.qp_temperatures: list[ndarray] = None  # type: ignore
        self.qp_heat_fluxes: list[ndarray] = None  # type: ignore
        self.qp_ddsddts: list[ndarray] = None  # type: ignore

    def update_element_material_stiffness_fint(self,
                                               is_update_material: bool = True,
                                               is_update_stiffness: bool = True,
                                               is_update_fint: bool = True, ) -> None:
        element_id = self.element_id
        timer = self.timer
        ntens = self.ntens
        ndi = self.ndi
        nshr = self.nshr

        qp_number = self.qp_number
        qp_weight_times_jacobi_dets = self.qp_weight_times_jacobi_dets
        qp_shape_values = self.iso_element_shape.qp_shape_values
        qp_dhdxex = self.qp_dhdxes

        qp_state_variables = self.qp_state_variables
        qp_state_variables_new = self.qp_state_variables_new

        element_dof_values = self.element_dof_values
        element_ddof_values = self.element_ddof_values

        material_data = self.material_data_list[0]

        if is_update_stiffness:
            self.element_stiffness = zeros(shape=(self.element_dof_number, self.element_dof_number), dtype=DTYPE)

        if is_update_fint:
            self.element_fint = zeros(self.element_dof_number, dtype=DTYPE)

        if is_update_material:
            self.qp_ddsddts = list()
            self.qp_temperatures = list()
            self.qp_heat_fluxes = list()

        for i in range(qp_number):
            if is_update_material:
                qp_weight_times_jacobi_det = qp_weight_times_jacobi_dets[i]
                qp_shape_value = qp_shape_values[i]
                qp_dhdx = qp_dhdxex[i]
                qp_temperature = dot(qp_shape_value, element_dof_values)
                qp_dtemperature = dot(qp_shape_value, element_ddof_values)
                qp_temperature_gradient = dot(qp_dhdx, element_dof_values)
                qp_dtemperature_gradient = dot(qp_dhdx, element_ddof_values)

                variable = {'temperature': qp_temperature,
                            'dtemperature': qp_dtemperature,
                            'temperature_gradient': qp_temperature_gradient,
                            'dtemperature_gradient': qp_dtemperature_gradient}
                qp_ddsddt, qp_output = material_data.get_tangent(variable=variable,
                                                                 state_variable=qp_state_variables[i],
                                                                 state_variable_new=qp_state_variables_new[i],
                                                                 element_id=element_id,
                                                                 iqp=i,
                                                                 ntens=ntens,
                                                                 ndi=ndi,
                                                                 nshr=nshr,
                                                                 timer=timer)
                qp_heat_flux = qp_output['heat_flux']
                self.qp_ddsddts.append(qp_ddsddt)
                self.qp_temperatures.append(qp_temperature)
                self.qp_heat_fluxes.append(qp_heat_flux)
            else:
                qp_weight_times_jacobi_det = qp_weight_times_jacobi_dets[i]
                qp_ddsddt = self.qp_ddsddts[i]
                qp_heat_flux = self.qp_heat_fluxes[i]

            if is_update_stiffness:
                self.element_stiffness += dot(qp_dhdx.transpose(), dot(qp_ddsddt, qp_dhdx)) * qp_weight_times_jacobi_det

            if is_update_fint:
                self.element_fint += dot(qp_dhdx.transpose(), qp_heat_flux) * qp_weight_times_jacobi_det

    def update_element_field_variables(self) -> None:
        self.qp_field_variables['heat_flux'] = array(self.qp_heat_fluxes, dtype=DTYPE)
        for key in self.qp_state_variables_new[0].keys():
            if key not in ['heat_flux']:
                variable = []
                for qp_state_variable_new in self.qp_state_variables_new:
                    variable.append(qp_state_variable_new[key])
                self.qp_field_variables[f'SDV-{key}'] = array(variable, dtype=DTYPE)
        self.element_nodal_field_variables = set_element_field_variables(self.qp_field_variables, self.iso_element_shape, self.dimension)


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(Thermal.__slots_dict__)

    from pyfem.Job import Job

    job = Job(r'F:\Github\pyfem\examples\thermal\1element\hex8\Job-1.toml')

    job.run()



# -*- coding: utf-8 -*-
"""

"""
from numpy import array, zeros, dot, ndarray

from pyfem.elements.BaseElement import BaseElement
from pyfem.elements.set_element_field_variables import set_element_field_variables
from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Dof import Dof
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.isoelements.IsoElementShape import IsoElementShape
from pyfem.materials.get_material_data import MaterialData
from pyfem.utils.colors import error_style


class SolidSmallStrain(BaseElement):
    """
    **固体小变形单元**

    :ivar qp_b_matrices: 积分点处的B矩阵列表
    :vartype qp_b_matrices: ndarray

    :ivar qp_b_matrices_transpose: 积分点处的B矩阵转置列表
    :vartype qp_b_matrices_transpose: ndarray

    :ivar qp_strains: 积分点处的应变列表
    :vartype qp_strains: list[ndarray]

    :ivar qp_stresses: 积分点处的应力列表
    :vartype qp_stresses: list[ndarray]

    :ivar ntens: 总应力数量
    :vartype ntens: int

    :ivar ndi: 轴向应力数量
    :vartype ndi: int

    :ivar nshr: 剪切应力数量
    :vartype nshr: int

    测试：比较当前代码和商业有限元软件 ABAQUS 的刚度矩阵，可以通过修改 ABAQUS inp文件，添加以下代码，将单元刚度矩阵输出到 ELEMENTSTIFFNESS.mtx 文件中::

        *Output, history, variable=PRESELECT
        *Element Matrix Output, Elset=Part-1-1.Set-All, File Name=ElementStiffness, Output File=User Defined, stiffness=yes

    我们可以发现 ABAQUS 使用的单元刚度矩阵和当前代码计算的刚度矩阵有一定的差别，这是由于 ABAQUS 采用了 B-Bar 方法对 B 矩阵进行了修正。

    注意：当前单元均为原始形式，存在剪切自锁，体积自锁，沙漏模式和零能模式等误差模式。几种误差模式的描述可以参考 https://blog.csdn.net/YORU_NO_KUNI/article/details/130370094。
    """

    __slots_dict__: dict = {
        'qp_b_matrices': ('ndarray', '积分点处的B矩阵列表'),
        'qp_b_matrices_transpose': ('ndarray', '积分点处的B矩阵转置列表'),
        'qp_strains': ('list[ndarray]', '积分点处的应变列表'),
        'qp_dstrains': ('list[ndarray]', '积分点处的应变增量列表'),
        'qp_stresses': ('list[ndarray]', '积分点处的应力列表'),
        'ntens': ('int', '总应力数量'),
        'ndi': ('int', '轴向应力数量'),
        'nshr': ('int', '剪切应力数量')
    }

    __slots__: list = BaseElement.__slots__ + [slot for slot in __slots_dict__.keys()]

    __allowed_material_data_list__ = [('ElasticIsotropic', 'PlasticKinematicHardening', 'PlasticCrystal', 'PlasticCrystalGNDs', 'ViscoElasticMaxwell', 'User')]

    def __init__(self, element_id: int,
                 iso_element_shape: IsoElementShape,
                 connectivity: ndarray,
                 node_coords: ndarray,
                 dof: Dof,
                 materials: list[Material],
                 section: Section,
                 material_data_list: list[MaterialData],
                 timer: Timer) -> None:

        super().__init__(element_id, iso_element_shape, connectivity, node_coords)

        self.allowed_material_data_list = self.__allowed_material_data_list__
        self.allowed_material_number = 1

        self.dof = dof
        self.materials = materials
        self.section = section
        self.material_data_list = material_data_list
        self.check_materials()
        self.timer = timer

        if self.dimension == 2:
            self.dof_names = ['u1', 'u2']
            self.ntens = 4
            self.ndi = 3
            self.nshr = 1
        elif self.dimension == 3:
            self.dof_names = ['u1', 'u2', 'u3']
            self.ntens = 6
            self.ndi = 3
            self.nshr = 3
        else:
            error_msg = f'{self.dimension} is not the supported dimension'
            raise NotImplementedError(error_style(error_msg))

        if dof.names != self.dof_names:
            error_msg = f'{dof.names} is not the supported dof of {type(self).__name__} element'
            raise NotImplementedError(error_style(error_msg))

        element_dof_number = len(self.dof_names) * self.iso_element_shape.nodes_number
        self.element_dof_number = element_dof_number
        self.element_dof_values = zeros(element_dof_number, dtype=DTYPE)
        self.element_ddof_values = zeros(element_dof_number, dtype=DTYPE)
        self.element_fint = zeros(element_dof_number, dtype=DTYPE)
        self.element_stiffness = zeros(shape=(self.element_dof_number, self.element_dof_number), dtype=DTYPE)

        self.qp_b_matrices: ndarray = None  # type: ignore
        self.qp_b_matrices_transpose: ndarray = None  # type: ignore
        self.qp_strains: list[ndarray] = None  # type: ignore
        self.qp_dstrains: list[ndarray] = None  # type: ignore
        self.qp_stresses: list[ndarray] = None  # type: ignore

        self.create_qp_b_matrices()

    def create_qp_b_matrices(self) -> None:
        if self.dimension == 2:
            self.qp_b_matrices = zeros(shape=(self.qp_number, 3, self.element_dof_number), dtype=DTYPE)
            for iqp, qp_dhdx in enumerate(self.qp_dhdxes):
                for i, val in enumerate(qp_dhdx.transpose()):
                    self.qp_b_matrices[iqp, 0, i * 2 + 0] = val[0]
                    self.qp_b_matrices[iqp, 1, i * 2 + 1] = val[1]
                    self.qp_b_matrices[iqp, 2, i * 2 + 0] = val[1]
                    self.qp_b_matrices[iqp, 2, i * 2 + 1] = val[0]

        elif self.dimension == 3:
            self.qp_b_matrices = zeros(shape=(self.iso_element_shape.qp_number, 6, self.element_dof_number), dtype=DTYPE)
            for iqp, qp_dhdx in enumerate(self.qp_dhdxes):
                for i, val in enumerate(qp_dhdx.transpose()):
                    self.qp_b_matrices[iqp, 0, i * 3 + 0] = val[0]
                    self.qp_b_matrices[iqp, 1, i * 3 + 1] = val[1]
                    self.qp_b_matrices[iqp, 2, i * 3 + 2] = val[2]
                    self.qp_b_matrices[iqp, 3, i * 3 + 0] = val[1]
                    self.qp_b_matrices[iqp, 3, i * 3 + 1] = val[0]
                    self.qp_b_matrices[iqp, 4, i * 3 + 0] = val[2]
                    self.qp_b_matrices[iqp, 4, i * 3 + 2] = val[0]
                    self.qp_b_matrices[iqp, 5, i * 3 + 1] = val[2]
                    self.qp_b_matrices[iqp, 5, i * 3 + 2] = val[1]

        self.qp_b_matrices_transpose = array([qp_b_matrix.transpose() for qp_b_matrix in self.qp_b_matrices])

    def update_element_material_stiffness_fint(self,
                                               is_update_material: bool = True,
                                               is_update_stiffness: bool = True,
                                               is_update_fint: bool = True, ) -> None:
        element_id = self.element_id
        timer = self.timer
        ntens = self.ntens
        ndi = self.ndi
        nshr = self.nshr

        qp_number = self.qp_number
        qp_b_matrices = self.qp_b_matrices
        qp_b_matrices_transpose = self.qp_b_matrices_transpose
        qp_weight_times_jacobi_dets = self.qp_weight_times_jacobi_dets

        qp_state_variables = self.qp_state_variables
        qp_state_variables_new = self.qp_state_variables_new

        element_dof_values = self.element_dof_values
        element_ddof_values = self.element_ddof_values

        material_data = self.material_data_list[0]

        if is_update_stiffness:
            self.element_stiffness = zeros(shape=(self.element_dof_number, self.element_dof_number), dtype=DTYPE)

        if is_update_fint:
            self.element_fint = zeros(self.element_dof_number, dtype=DTYPE)

        if is_update_material:
            self.qp_ddsddes = list()
            self.qp_strains = list()
            self.qp_dstrains = list()
            self.qp_stresses = list()

        for i in range(qp_number):
            if is_update_material:
                qp_weight_times_jacobi_det = qp_weight_times_jacobi_dets[i]
                qp_b_matrix_transpose = qp_b_matrices_transpose[i]
                qp_b_matrix = qp_b_matrices[i]
                qp_strain = dot(qp_b_matrix, element_dof_values)
                qp_dstrain = dot(qp_b_matrix, element_ddof_values)
                variable = {'strain': qp_strain, 'dstrain': qp_dstrain}
                qp_ddsdde, qp_output = material_data.get_tangent(variable=variable,
                                                                 state_variable=qp_state_variables[i],
                                                                 state_variable_new=qp_state_variables_new[i],
                                                                 element_id=element_id,
                                                                 iqp=i,
                                                                 ntens=ntens,
                                                                 ndi=ndi,
                                                                 nshr=nshr,
                                                                 timer=timer)
                qp_stress = qp_output['stress']
                self.qp_ddsddes.append(qp_ddsdde)
                self.qp_strains.append(qp_strain)
                self.qp_dstrains.append(qp_dstrain)
                self.qp_stresses.append(qp_stress)
            else:
                qp_b_matrix_transpose = qp_b_matrices_transpose[i]
                qp_b_matrix = qp_b_matrices[i]
                qp_weight_times_jacobi_det = qp_weight_times_jacobi_dets[i]
                qp_ddsdde = self.qp_ddsddes[i]
                qp_stress = self.qp_stresses[i]

            if is_update_stiffness:
                self.element_stiffness += dot(qp_b_matrix_transpose, dot(qp_ddsdde, qp_b_matrix)) * qp_weight_times_jacobi_det

            if is_update_fint:
                self.element_fint += dot(qp_b_matrix_transpose, qp_stress) * qp_weight_times_jacobi_det

    def update_element_field_variables(self) -> None:
        self.qp_field_variables['strain'] = array(self.qp_strains, dtype=DTYPE) + array(self.qp_dstrains, dtype=DTYPE)
        self.qp_field_variables['stress'] = array(self.qp_stresses, dtype=DTYPE)
        for key in self.qp_state_variables_new[0].keys():
            if key not in ['strain', 'stress']:
                variable = []
                for qp_state_variable_new in self.qp_state_variables_new:
                    variable.append(qp_state_variable_new[key])
                self.qp_field_variables[f'SDV-{key}'] = array(variable, dtype=DTYPE)
        self.element_nodal_field_variables = set_element_field_variables(self.qp_field_variables, self.iso_element_shape, self.dimension)


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(SolidSmallStrain.__slots_dict__)

    # from pyfem.Job import Job
    #
    # job = Job(r'..\..\..\examples\mechanical\1element\hex8\Job-1.toml')
    #
    # job.assembly.element_data_list[0].show()



# -*- coding: utf-8 -*-
"""

"""
from copy import deepcopy

from numpy import eye, ndarray, dot, zeros

from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style


class ThermalIsotropic(BaseMaterial):
    """
    各项同性热传导材料。

    支持的截面属性：('', 'Volume', 'PlaneStress', 'PlaneStrain')

    :ivar k: Conductivity k
    :vartype k: float

    :ivar cp: Capacity cp
    :vartype cp: float
    """

    __slots_dict__: dict = {
        'k': ('float', 'Conductivity k'),
        'cp': ('float', 'Capacity cp'),
    }

    __slots__ = BaseMaterial.__slots__ + [slot for slot in __slots_dict__.keys()]

    __data_keys__ = ['Conductivity k', 'Capacity cp']

    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        super().__init__(material, dimension, section)
        self.allowed_section_types = ('', 'Volume', 'PlaneStress', 'PlaneStrain')

        self.data_keys = self.__data_keys__

        if len(self.material.data) != len(self.data_keys):
            raise NotImplementedError(error_style(self.get_data_length_error_msg()))
        else:
            for i, key in enumerate(self.data_keys):
                self.data_dict[key] = material.data[i]

        self.k: float = self.data_dict['Conductivity k']
        self.cp: float = self.data_dict['Capacity cp']

        self.create_tangent()

    def create_tangent(self):
        if self.section.type in self.allowed_section_types:
            self.tangent = eye(self.dimension) * self.k
        else:
            raise NotImplementedError(error_style(self.get_section_type_error_msg()))

    def get_tangent(self, variable: dict[str, ndarray],
                    state_variable: dict[str, ndarray],
                    state_variable_new: dict[str, ndarray],
                    element_id: int,
                    iqp: int,
                    ntens: int,
                    ndi: int,
                    nshr: int,
                    timer: Timer) -> tuple[ndarray, dict[str, ndarray]]:
        # 全量格式
        # temperature_gradient = variable['temperature_gradient']
        # heat_flux = dot(-self.tangent, temperature_gradient)

        # 增量格式
        temperature_gradient = variable['temperature_gradient']
        dtemperature_gradient = variable['dtemperature_gradient']
        if state_variable == {} or timer.time0 == 0.0:
            state_variable['heat_flux'] = zeros(len(temperature_gradient), dtype=DTYPE)
        heat_flux = deepcopy(state_variable['heat_flux'])
        heat_flux += dot(-self.tangent, dtemperature_gradient)
        state_variable_new['heat_flux'] = heat_flux

        output = {'heat_flux': heat_flux}
        return self.tangent, output


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(ThermalIsotropic.__slots_dict__)

    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'..\..\..\examples\thermal\1element\hex8\Job-1.toml')
    material_data = ThermalIsotropic(props.materials[0], 3, props.sections[0])
    material_data.show()



# -*- coding: utf-8 -*-
"""

"""
import os

import h5py  # type: ignore
from numpy import array, ndarray

from pyfem import __version__
from pyfem.assembly.Assembly import Assembly
from pyfem.utils.visualization import object_slots_to_string_ndarray


class Database:
    """
    求解器基类。

    :ivar assembly: 装配体对象
    :vartype assembly: Assembly

    :ivar solver: 求解器属性
    :vartype solver: Solver

    :ivar dof_solution: 求解得到自由度的值
    :vartype dof_solution: ndarray
    """

    __slots_dict__: dict = {
        'assembly': ('Assembly', '装配体对象'),
        'title': ('str', '装配体对象'),
        'version': ('str', '求解器属性'),
        'materials': ('dict', '求解器属性'),
        'dimension': ('ndarray', '求解得到自由度的值'),
        'nodes': ('ndarray', '求解得到自由度的值'),
        'elements': ('ndarray', '求解得到自由度的值'),
        'node_sets': ('ndarray', '求解得到自由度的值'),
        'element_sets': ('ndarray', '求解得到自由度的值'),
        'steps': ('ndarray', '求解得到自由度的值'),
    }

    __slots__: list = [slot for slot in __slots_dict__.keys()]

    def __init__(self, assembly: Assembly) -> None:
        self.assembly: Assembly = assembly
        self.title: str = assembly.props.title
        self.version: str = f'PYFEM {__version__}'
        self.materials: dict = {}
        self.dimension: int = assembly.props.mesh_data.dimension
        self.nodes: ndarray = assembly.props.mesh_data.nodes
        self.elements: list[ndarray] = assembly.props.mesh_data.elements
        self.node_sets: dict[str, list[int]] = assembly.props.mesh_data.node_sets
        self.element_sets: dict[str, list[int]] = assembly.props.mesh_data.element_sets
        self.steps: dict = {}

    def to_string(self, level: int = 1) -> str:
        return object_slots_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def write(self) -> None:
        for output in self.assembly.props.outputs:
            if output.is_save:
                if output.type == 'vtk':
                    pass
                if output.type == 'hdf5':
                    self.add_hdf5()

    def init_hdf5(self) -> None:
        job_name = self.assembly.props.input_file.stem
        output_file = self.assembly.props.work_path.joinpath(f'{job_name}.hdf5')
        if os.path.exists(output_file):
            os.remove(output_file)
        self.add_hdf5()

    def add_hdf5(self) -> None:
        assembly = self.assembly
        props = self.assembly.props
        timer = self.assembly.timer

        job_name = props.input_file.stem
        output_file = props.work_path.joinpath(f'{job_name}.hdf5')

        f = h5py.File(output_file, 'a')

        if 'title' not in f.keys():
            f.create_dataset('title', data=self.title)
        if 'version' not in f.keys():
            f.create_dataset('version', data=self.version)
        if 'dimension' not in f.keys():
            f.create_dataset('dimension', data=self.dimension)
        if 'nodes' not in f.keys():
            f.create_dataset('nodes', data=self.nodes)
        if 'elements' not in f.keys():
            elements_group = f.create_group('elements')

            connectivity_list = []
            cells_list = []
            offset_list = [0]
            celltypes_list = []

            offset = 0
            for connectivity in self.elements:
                cells_list.append(len(connectivity))
                for node_id in connectivity:
                    connectivity_list.append(node_id)
                    cells_list.append(node_id)
                offset += len(connectivity)
                offset_list.append(offset)
                if self.dimension == 2:
                    celltypes_list.append(9)
                elif self.dimension == 3:
                    celltypes_list.append(12)

            elements_group.create_dataset('connectivity', data=array(connectivity_list, dtype='int32'))
            elements_group.create_dataset('offsets', data=array(offset_list, dtype='int32'))
            elements_group.create_dataset('cells', data=array(cells_list, dtype='int32'))
            elements_group.create_dataset('celltypes', data=array(celltypes_list, dtype='int32'))

        if 'node_sets' not in f.keys():
            f.create_group('node_sets')
            for key, value in self.node_sets.items():
                f['node_sets'].create_dataset(key, data=value)

        if 'element_sets' not in f.keys():
            f.create_group('element_sets')
            for key, value in self.element_sets.items():
                f['element_sets'].create_dataset(key, data=value)

        if 'steps' not in f.keys():
            f.create_group('steps')
            for step_name in ['Step-1']:
                f['steps'].create_group(step_name)
                frameId = str(0)
                f['steps'][step_name].create_dataset('frame_count', data=int(frameId))
                f['steps'][step_name].create_group('frames')
        else:
            for step_name in ['Step-1']:
                frame_count = f['steps'][step_name]['frame_count'][()]
                frameId = str(frame_count + 1)
                frame_count = f['steps'][step_name]['frame_count'][()] = int(frameId)

        for step_name in ['Step-1']:
            f['steps'][step_name]['frames'].create_group(frameId)
            f['steps'][step_name]['frames'][frameId].create_dataset('incrementNumber', data=timer.increment)
            f['steps'][step_name]['frames'][frameId].create_dataset('frameValue', data=timer.time1)
            f['steps'][step_name]['frames'][frameId].create_group('fieldOutputs')

            if 'T' in props.dof.names:
                col_T = props.dof.names.index('T')
                dof_T = assembly.dof_solution.reshape(-1, len(props.dof.names))[:, col_T]
                f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group('T')
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['T'].create_dataset('bulkDataBlocks', data=dof_T)
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['T'].create_dataset('componentLabels', data=array([''], dtype=object))
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['T'].create_dataset('validInvariants', data=array([''], dtype=object))
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['T'].create_dataset('description', data='Temperature')
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['T'].create_dataset('type', data='SCALAR')

            if 'u1' in props.dof.names:
                if self.dimension == 2:
                    data_value = assembly.dof_solution.reshape(-1, len(props.dof.names))[:, 0:2]
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group('U')
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['U'].create_dataset('bulkDataBlocks', data=data_value)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['U'].create_dataset('componentLabels', data=array(['U1', 'U2'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['U'].create_dataset('validInvariants', data=array(['Magnitude'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['U'].create_dataset('description', data='Displacement')
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['U'].create_dataset('type', data='VECTOR')
                elif self.dimension == 3:
                    data_value = assembly.dof_solution.reshape(-1, len(props.dof.names))[:, 0:3]
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group('U')
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['U'].create_dataset('bulkDataBlocks', data=data_value)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['U'].create_dataset('componentLabels',
                                                                                                 data=array(['U1', 'U2', 'U3'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['U'].create_dataset('validInvariants', data=array(['Magnitude'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['U'].create_dataset('description', data='Displacement')
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['U'].create_dataset('type', data='VECTOR')
                else:
                    raise NotImplementedError

                if self.dimension == 2:
                    data_value = assembly.fint.reshape(-1, len(props.dof.names))[:, 0:2]
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group('RF')
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['RF'].create_dataset('bulkDataBlocks', data=data_value)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['RF'].create_dataset('componentLabels', data=array(['RF1', 'RF2'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['RF'].create_dataset('validInvariants', data=array(['Magnitude'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['RF'].create_dataset('description', data='Displacement')
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['RF'].create_dataset('type', data='VECTOR')
                elif self.dimension == 3:
                    data_value = assembly.fint.reshape(-1, len(props.dof.names))[:, 0:3]
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group('RF')
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['RF'].create_dataset('bulkDataBlocks', data=data_value)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['RF'].create_dataset('componentLabels',
                                                                                                  data=array(['RF1', 'RF2', 'RF3'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['RF'].create_dataset('validInvariants', data=array(['Magnitude'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['RF'].create_dataset('description', data='Displacement')
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs']['RF'].create_dataset('type', data='VECTOR')
                else:
                    raise NotImplementedError

            if "phi" in props.dof.names:
                col_phi = props.dof.names.index("phi")
                dof_phi = assembly.dof_solution.reshape(-1, len(props.dof.names))[:, col_phi]
                f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group('PHI')
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['PHI'].create_dataset('bulkDataBlocks', data=dof_phi)
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['PHI'].create_dataset('componentLabels', data=())
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['PHI'].create_dataset('validInvariants', data=())
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['PHI'].create_dataset('description', data='Phasefield')
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['PHI'].create_dataset('type', data='SCALAR')

            if "C" in props.dof.names:
                col_c = props.dof.names.index("C")
                dof_c = assembly.dof_solution.reshape(-1, len(props.dof.names))[:, col_c]
                f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group('C')
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['C'].create_dataset('bulkDataBlocks', data=dof_c)
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['C'].create_dataset('componentLabels', data=())
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['C'].create_dataset('validInvariants', data=())
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['C'].create_dataset('description', data='Concentration')
                f['steps'][step_name]['frames'][frameId]['fieldOutputs']['C'].create_dataset('type', data='SCALAR')

            for key, data_value in assembly.field_variables.items():
                if key == 'E':
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group(key)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('bulkDataBlocks', data=data_value)
                    if self.dimension == 2:
                        f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('componentLabels', data=array(['E11', 'E22', 'E12'], dtype=object))
                    if self.dimension == 3:
                        f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('componentLabels', data=array(['E11', 'E22', 'E33', 'E12', 'E13', 'E23'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('validInvariants', data=array(['Max. Principal'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('description', data=key)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('type', data='TENSOR')
                elif key == 'S':
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group(key)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('bulkDataBlocks', data=data_value)
                    if self.dimension == 2:
                        f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('componentLabels', data=array(['S11', 'S22', 'S12'], dtype=object))
                    if self.dimension == 3:
                        f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('componentLabels', data=array(['S11', 'S22', 'S33', 'S12', 'S13', 'S23'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('validInvariants', data=array(['Mises'], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('description', data=key)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('type', data='TENSOR')
                elif 'SDV' in key:
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group(key)
                    if len(data_value.shape) == 1:
                        f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('bulkDataBlocks', data=data_value)
                        f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('componentLabels', data=())
                    if len(data_value.shape) == 2:
                        f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('bulkDataBlocks', data=data_value)
                        f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('componentLabels', data=array([f'{i+1}' for i in range(data_value.shape[1])], dtype=object))
                    if len(data_value.shape) == 3:
                        f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('bulkDataBlocks', data=data_value.reshape(data_value.shape[0], -1))
                        labels_str = []
                        for i in range(data_value.shape[1]):
                            for j in range(data_value.shape[2]):
                                labels_str.append(f'{i+1},{j+1}')
                        f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('componentLabels', data=array(labels_str, dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('validInvariants', data=array([''], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('description', data=key)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('type', data='SCALAR')
                else:
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'].create_group(key)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('bulkDataBlocks', data=data_value)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('componentLabels', data=array([''], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('validInvariants', data=array([''], dtype=object))
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('description', data=key)
                    f['steps'][step_name]['frames'][frameId]['fieldOutputs'][key].create_dataset('type', data='SCALAR')

            f['steps'][step_name]['frames'][frameId].create_group('solution')
            f['steps'][step_name]['frames'][frameId]['solution'].create_dataset('dof_solution', data=self.assembly.dof_solution)
            f['steps'][step_name]['frames'][frameId]['solution'].create_dataset('fint', data=self.assembly.fint)
            f['steps'][step_name]['frames'][frameId]['solution'].create_dataset('fext', data=self.assembly.fext)

        f.close()


if __name__ == "__main__":
    # print_slots_dict(Database.__slots_dict__)

    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'..\..\..\examples\mechanical\plane\Job-1.toml')
    assembly = Assembly(props)

    database = Database(assembly)

    database.init_hdf5()
    database.add_hdf5()
    # print(database.elements)

    # database.show()



# -*- coding: utf-8 -*-
"""

"""
import h5py  # type: ignore
from numpy import array, ndarray, empty, zeros, hstack
from numpy.linalg import norm

from pyfem.utils.visualization import object_slots_to_string


class ODB:
    """
    cells = [4, 2, 6, 8, 5, 4, 5, 8, 4, 1, 4, 6, 3, 7, 8, 4, 8, 7, 0, 4]
    elements = [[2, 6, 8, 5], [5, 8, 4, 1], [6, 3, 7, 8], [8, 7, 0, 4]]
    """

    __slots_dict__: dict = {
        'title': ('str', 'odb名称'),
        'path': ('str', '文件路径'),
        'version': ('str', '求解器版本'),
        'materials': ('dict', '材料字典'),
        'dimension': ('int', '空间维度'),
        'nodes': ('ndarray', '节点坐标数组'),
        'cells': ('ndarray', '单元连接信息数组'),
        'celltypes': ('ndarray', '单元类型数组'),
        'elements': ('list[ndarray]', '单元connectivity列表'),
        'node_sets': ('dict[str, list[int]]', '节点集合字典'),
        'element_sets': ('dict[str, list[int]]', '单元集合字典'),
        'steps': ('dict', '载荷步字典'),
        'vtu_count': ('int', '读取vtu文件时的计数器'),
        'is_deformed': ('bool', '判断数据中是否有位移'),
        'field_output_labels': ('dict', 'fieldOutputs变量的标签字典'),
        'xy_data_dict': ('dict', '存储导出的xy数据字典'),
        'current_step': ('str', 'interactor中当前正在渲染的step名字'),
    }

    __slots__: list = [slot for slot in __slots_dict__.keys()]

    def __init__(self) -> None:
        self.title: str = ''
        self.path: str = ''
        self.version: str = ''
        self.materials: dict = {}
        self.dimension: int = -1
        self.nodes: ndarray = empty(0)
        self.cells: ndarray = empty(0)
        self.celltypes: ndarray = empty(0)
        self.elements: list[ndarray] = []
        self.node_sets: dict[str, list[int]] = {}
        self.element_sets: dict[str, list[int]] = {}
        self.steps: dict = {}
        self.vtu_count: int = 0
        self.is_deformed: bool = False
        self.field_output_labels: dict = {}
        self.xy_data_dict: dict[str, ndarray] = {}
        self.current_step = 'Step-1'

    def reset(self):
        self.__init__()

    def load_hdf5(self, file_path: str) -> str:
        try:
            f = h5py.File(file_path, 'r')
            self.title = f['title'][()]
            self.path = file_path
            self.version = f['version'][()]
            self.dimension = f['dimension'][()]
            self.nodes = f['nodes'][()]

            if self.dimension == 2:
                new_column = zeros((self.nodes.shape[0], 1))
                self.nodes = hstack((self.nodes, new_column))
            self.cells = f['elements']['cells'][()]
            self.celltypes = f['elements']['celltypes'][()]

            # 下面的函数用于从cells计算单元的connectivity列表
            i = 0
            while i < len(self.cells):
                n = self.cells[i]
                self.elements.append(self.cells[i + 1:i + 1 + n])
                i += n + 1

            for key, value in f['node_sets'].items():
                self.node_sets[key] = value[()]

            for key, value in f['element_sets'].items():
                self.element_sets[key] = value[()]

            for step_name in f['steps'].keys():
                self.steps[step_name] = {
                    'frames': []
                }
                self.field_output_labels[step_name] = {}

                frame_ids = sorted([int(frame_id) for frame_id in f['steps'][step_name]['frames'].keys()])

                for frame_id in frame_ids:
                    frame_data = {'incrementNumber': f['steps'][step_name]['frames'][str(frame_id)]['incrementNumber'][()],
                                  'frameValue': f['steps'][step_name]['frames'][str(frame_id)]['frameValue'][()],
                                  'fieldOutputs': {},
                                  'solution': {}}
                    for key, value in f['steps'][step_name]['frames'][str(frame_id)]['fieldOutputs'].items():
                        frame_data['fieldOutputs'][key] = {}
                        if self.dimension == 2 and key == 'U':
                            u_2d = value['bulkDataBlocks'][()]
                            u_3d = hstack((u_2d, zeros((u_2d.shape[0], 1))))
                            frame_data['fieldOutputs'][key]['bulkDataBlocks'] = u_3d
                        else:
                            frame_data['fieldOutputs'][key]['bulkDataBlocks'] = value['bulkDataBlocks'][()]

                        componentLabels = [v.decode() for v in value['componentLabels'][()] if v != b'']
                        validInvariants = [v.decode() for v in value['validInvariants'][()] if v != b'']
                        description = value['description'][()].decode()
                        type = value['type'][()].decode()

                        frame_data['fieldOutputs'][key]['componentLabels'] = componentLabels
                        frame_data['fieldOutputs'][key]['validInvariants'] = validInvariants
                        frame_data['fieldOutputs'][key]['description'] = description
                        frame_data['fieldOutputs'][key]['type'] = type

                        self.field_output_labels[step_name][key] = {'componentLabels': componentLabels,
                                                                    'validInvariants': validInvariants,
                                                                    'description': description,
                                                                    'type': type}
                    for key, value in f['steps'][step_name]['frames'][str(frame_id)]['solution'].items():
                        frame_data['solution'][key] = value[()]

                    self.steps[step_name]['frames'].append(frame_data)

            f.close()
            return 'Succeed to load file: ' + f'\"{file_path}\"\n'
        except NotImplementedError:
            try:
                f.close()
            except NotImplementedError:
                pass
            return 'Failed to load file: ' + f'\"{file_path}\"\n'

    def get_nodes_data(self, step_name: str, nodes: list[int], name: str) -> tuple[ndarray, ndarray, ndarray]:
        field_name = name
        frame_ids = []
        frame_values = []
        data = []
        frames = self.steps[step_name]['frames']
        for frame_id, frame_value in enumerate(frames):
            frame_ids.append(frame_id)
            frame_values.append(frame_value['frameValue'])
            data.append(frame_value['fieldOutputs'][field_name]['bulkDataBlocks'][nodes])
        return array(frame_ids), array(frame_values), array(data)

    def get_node_set_data(self, step_name: str, node_set_name: str, name: str) -> tuple[ndarray, ndarray, ndarray]:
        return self.get_nodes_data(step_name, self.node_sets[node_set_name], name)

    def to_string(self, level: int = 1) -> str:
        return object_slots_to_string(self, level)

    def show(self) -> None:
        print(self.to_string())

    def get_frame_data(self, step_name: str, frame_id: int, name: str) -> ndarray:
        if len(name.split('-')) == 1:
            field_name = name
            field_data = self.steps[step_name]['frames'][frame_id]['fieldOutputs'][field_name]
            return field_data['bulkDataBlocks']

        elif len(name.split('-')) == 2:
            field_name = name.split('-')[0]
            field_option = name.split('-')[1]
            field_data = self.steps[step_name]['frames'][frame_id]['fieldOutputs'][field_name]

            if field_option in field_data['componentLabels']:
                col = field_data['componentLabels'].index(field_option)
                return field_data['bulkDataBlocks'][:, col]
            if field_option == 'Magnitude':
                return norm(field_data['bulkDataBlocks'], axis=1)

        else:
            raise NotImplementedError


if __name__ == '__main__':
    odb = ODB()
    odb.load_hdf5(r'F:\Github\pyfem\examples\mechanical\rectangle_hole_3D\Job-1.hdf5')
    print(odb.field_output_labels)



# -*- coding: utf-8 -*-
"""

"""
from typing import Union

from pyfem.assembly.Assembly import Assembly
from pyfem.io.Solver import Solver
from pyfem.solvers.ArcLengthSolver import ArcLengthSolver
from pyfem.solvers.BaseSolver import BaseSolver
from pyfem.solvers.LinearSolver import LinearSolver
from pyfem.solvers.NonlinearSolver import NonlinearSolver
from pyfem.solvers.TimeIntegrationNonlinearSolver import TimeIntegrationNonlinearSolver
from pyfem.utils.colors import error_style

SolverData = Union[BaseSolver, LinearSolver, NonlinearSolver, ArcLengthSolver]

solver_data_dict = {
    'LinearSolver': LinearSolver,
    'NonlinearSolver': NonlinearSolver,
    'ArcLengthSolver': ArcLengthSolver,
    'TimeIntegrationNonlinearSolver': TimeIntegrationNonlinearSolver,
}


def get_solver_data(assembly: Assembly, solver: Solver) -> SolverData:
    """
    工厂函数，用于根据求解器属性生产不同的求解器对象。

    Args:
        assembly(Assembly): 装配体对象
        solver(Solver): 求解器属性

    :return: 求解器对象
    :rtype: SolverData
    """

    class_name = f'{solver.type}'.strip().replace(' ', '')

    if class_name in solver_data_dict:
        return solver_data_dict[class_name](assembly=assembly,
                                            solver=solver)
    else:
        error_msg = f'{class_name} solver is not supported.\n'
        error_msg += f'The allowed solver types are {list(solver_data_dict.keys())}.'
        raise NotImplementedError(error_style(error_msg))


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style


class PhaseFieldDamageCZM(BaseMaterial):
    """
    相场断裂材料。

    支持的截面属性：('Volume', 'PlaneStress', 'PlaneStrain')

    :ivar gc: surface energy to create a unit fracture surface gc
    :vartype gc: float

    :ivar lc: length scale parameter to measure the damage diffusion lc
    :vartype lc: float
    """

    __slots_dict__: dict = {
        'gc': ('float', 'surface energy to create a unit fracture surface gc'),
        'lc': ('float', 'length scale parameter to measure the damage diffusion lc'),
        'a1': ('float', 'degradation function coefficient a1'),
        'a2': ('float', 'degradation function coefficient a2'),
        'a3': ('float', 'degradation function coefficient a3'),
        'p': ('float', 'polynomial power p'),
        'xi': ('float', 'geometric function coefficient xi'),
        'c0': ('float', 'band width c0'),
        'gth': ('float', 'surface energy threshold gth')
    }

    __slots__ = BaseMaterial.__slots__ + [slot for slot in __slots_dict__.keys()]

    __data_keys__ = ['surface energy to create a unit fracture surface gc',
                     'length scale parameter to measure the damage diffusion lc',
                     'degradation function coefficient a1',
                     'degradation function coefficient a2',
                     'degradation function coefficient a3',
                     'polynomial power p',
                     'geometric function coefficient xi',
                     'band width c0',
                     'surface energy threshold gth']

    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        super().__init__(material, dimension, section)
        self.allowed_section_types = ('Volume', 'PlaneStress', 'PlaneStrain')

        self.data_keys = self.__data_keys__

        if len(self.material.data) != len(self.data_keys):
            raise NotImplementedError(error_style(self.get_data_length_error_msg()))
        else:
            for i, key in enumerate(self.data_keys):
                self.data_dict[key] = material.data[i]

        self.gc: float = self.data_dict['surface energy to create a unit fracture surface gc']
        self.lc: float = self.data_dict['length scale parameter to measure the damage diffusion lc']
        self.a1: float = self.data_dict['degradation function coefficient a1']
        self.a2: float = self.data_dict['degradation function coefficient a2']
        self.a3: float = self.data_dict['degradation function coefficient a3']
        self.p: float = self.data_dict['polynomial power p']
        self.xi: float = self.data_dict['geometric function coefficient xi']
        self.c0: float = self.data_dict['band width c0']
        self.gth: float = self.data_dict['surface energy threshold gth']

        self.create_tangent()


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(PhaseFieldDamageCZM.__slots_dict__)



# -*- coding: utf-8 -*-
"""

"""
from numpy import ndarray, average, dot, tile


def set_element_field_variables(qp_field_variables: dict[str, ndarray], iso_element_shape, dimension: int) -> dict[str, ndarray]:
    """

    """
    method = ''

    element_nodal_values = {}
    for key, item in qp_field_variables.items():
        if iso_element_shape.extrapolated_matrix.shape != (0,) and method == 'extrapolated':
            if len(item.shape) == 1:
                element_nodal_values[key] = dot(iso_element_shape.extrapolated_matrix, item.reshape(-1, 1))
            else:
                element_nodal_values[key] = dot(iso_element_shape.extrapolated_matrix, item)
        else:
            if len(item.shape) == 1:
                tile_shape = (iso_element_shape.nodes_number, 1)
            elif len(item.shape) == 2:
                tile_shape = (iso_element_shape.nodes_number, 1)
            elif len(item.shape) == 3:
                tile_shape = (iso_element_shape.nodes_number, 1, 1)
            else:
                raise ValueError('The shape of field variable is not correct.')
            element_nodal_values[key] = tile(average(item, axis=0), tile_shape)

    element_nodal_field_variables = {}
    for key, item in element_nodal_values.items():
        if key == 'stress':
            if dimension == 2:
                element_nodal_field_variables['S11'] = item[:, 0]
                element_nodal_field_variables['S22'] = item[:, 1]
                element_nodal_field_variables['S12'] = item[:, 2]
            elif dimension == 3:
                element_nodal_field_variables['S11'] = item[:, 0]
                element_nodal_field_variables['S22'] = item[:, 1]
                element_nodal_field_variables['S33'] = item[:, 2]
                element_nodal_field_variables['S12'] = item[:, 3]
                element_nodal_field_variables['S13'] = item[:, 4]
                element_nodal_field_variables['S23'] = item[:, 5]
            element_nodal_field_variables['S'] = item

        if key == 'strain':
            if dimension == 2:
                element_nodal_field_variables['E11'] = item[:, 0]
                element_nodal_field_variables['E22'] = item[:, 1]
                element_nodal_field_variables['E12'] = item[:, 2]
            elif dimension == 3:
                element_nodal_field_variables['E11'] = item[:, 0]
                element_nodal_field_variables['E22'] = item[:, 1]
                element_nodal_field_variables['E33'] = item[:, 2]
                element_nodal_field_variables['E12'] = item[:, 3]
                element_nodal_field_variables['E13'] = item[:, 4]
                element_nodal_field_variables['E23'] = item[:, 5]
            element_nodal_field_variables['E'] = item

        if key == 'energy':
            element_nodal_field_variables['Energy'] = item[:, 0]

        if key == 'heat_flux':
            if dimension >= 1:
                element_nodal_field_variables['HFL1'] = item[:, 0]
            if dimension >= 2:
                element_nodal_field_variables['HFL2'] = item[:, 1]
            if dimension >= 3:
                element_nodal_field_variables['HFL3'] = item[:, 2]
            element_nodal_field_variables['HFL'] = item

        if key == 'concentration_flux':
            if dimension >= 1:
                element_nodal_field_variables['CFL1'] = item[:, 0]
            if dimension >= 2:
                element_nodal_field_variables['CFL2'] = item[:, 1]
            if dimension >= 3:
                element_nodal_field_variables['CFL3'] = item[:, 2]
            element_nodal_field_variables['CFL'] = item

        if 'SDV' in key:
            element_nodal_field_variables[key] = item

    return element_nodal_field_variables


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from copy import deepcopy

from numpy import eye, ndarray, dot, zeros

from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style


class DiffusionIsotropic(BaseMaterial):
    """
    各项同性热传导材料。

    支持的截面属性：('', 'Volume', 'PlaneStress', 'PlaneStrain')

    :ivar d: Diffusion coefficient d
    :vartype d: float
    """

    __slots_dict__: dict = {
        'd': ('float', 'Diffusion coefficient d'),
    }

    __slots__ = BaseMaterial.__slots__ + [slot for slot in __slots_dict__.keys()]

    __data_keys__ = ['Diffusion coefficient d']

    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        super().__init__(material, dimension, section)
        self.allowed_section_types = ('', 'Volume', 'PlaneStress', 'PlaneStrain')

        self.data_keys = self.__data_keys__

        if len(self.material.data) != len(self.data_keys):
            raise NotImplementedError(error_style(self.get_data_length_error_msg()))
        else:
            for i, key in enumerate(self.data_keys):
                self.data_dict[key] = material.data[i]

        self.d: float = self.data_dict['Diffusion coefficient d']

        self.create_tangent()

    def create_tangent(self):
        if self.section.type in self.allowed_section_types:
            self.tangent = eye(self.dimension) * self.d
        else:
            raise NotImplementedError(error_style(self.get_section_type_error_msg()))

    def get_tangent(self, variable: dict[str, ndarray],
                    state_variable: dict[str, ndarray],
                    state_variable_new: dict[str, ndarray],
                    element_id: int,
                    iqp: int,
                    ntens: int,
                    ndi: int,
                    nshr: int,
                    timer: Timer) -> tuple[ndarray, dict[str, ndarray]]:

        concentration_gradient = variable['concentration_gradient']
        dconcentration_gradient = variable['dconcentration_gradient']
        if state_variable == {} or timer.time0 == 0.0:
            state_variable['concentration_flux'] = zeros(len(concentration_gradient), dtype=DTYPE)
        concentration_flux = deepcopy(state_variable['concentration_flux'])
        concentration_flux += dot(-self.tangent, dconcentration_gradient)
        state_variable_new['concentration_flux'] = concentration_flux
        output = {'concentration_flux': concentration_flux}
        return self.tangent, output


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(DiffusionIsotropic.__slots_dict__)

    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'..\..\..\examples\diffusion\rectangle\Job-1.toml')
    material_data = DiffusionIsotropic(props.materials[0], 2, props.sections[0])
    material_data.show()



# -*- coding: utf-8 -*-
"""

"""
import numpy as np
from numpy import zeros, ndarray, array, sum, dot, maximum, tensordot
from numpy.linalg import inv, norm, eig

from pyfem.fem.constants import DTYPE
from pyfem.utils.colors import error_style


def inverse(qp_jacobis: ndarray, qp_jacobi_dets: ndarray) -> ndarray:
    r"""
    **求逆矩阵**

    :param qp_jacobis: 积分点处的雅克比矩阵列表
    :type qp_jacobis: ndarray

    :param qp_jacobi_dets: 积分点处的雅克比矩阵行列式列表
    :type qp_jacobi_dets: ndarray

    :return: 逆矩阵列表
    :rtype: ndarray

    当输入为 2×2 和 3×3 的矩阵直接通过解析式计算，其余的情况返回 :py:meth:`numpy.linalg.inv()` 函数的计算结果。

    对于 2×2 的矩阵：

    .. math::
        {\mathbf{A}} = \left[ {\begin{array}{*{20}{c}}
          {{A_{11}}}&{{A_{12}}} \\
          {{A_{21}}}&{{A_{22}}}
        \end{array}} \right]

    .. math::
        {{\mathbf{A}}^{ - 1}} = \frac{1}{{\det \left( {\mathbf{A}} \right)}}\left[ {\begin{array}{*{20}{c}}
          {{A_{22}}}&{ - {A_{12}}} \\
          { - {A_{21}}}&{{A_{11}}}
        \end{array}} \right]

    对于 3×3 的矩阵：

    .. math::
        {\mathbf{A}} = \left[ {\begin{array}{*{20}{c}}
          {{A_{11}}}&{{A_{12}}}&{{A_{13}}} \\
          {{A_{21}}}&{{A_{22}}}&{{A_{23}}} \\
          {{A_{31}}}&{{A_{32}}}&{{A_{33}}}
        \end{array}} \right]

    .. math::
        {{\mathbf{A}}^{ - 1}} = \frac{1}{{\det \left( {\mathbf{A}} \right)}}\left[ {\begin{array}{*{20}{c}}
          {{A_{22}}{A_{33}} - {A_{23}}{A_{32}}}&{{A_{13}}{A_{32}} - {A_{12}}{A_{33}}}&{{A_{12}}{A_{23}} - {A_{13}}{A_{22}}} \\
          {{A_{23}}{A_{31}} - {A_{21}}{A_{33}}}&{{A_{11}}{A_{33}} - {A_{13}}{A_{31}}}&{{A_{13}}{A_{21}} - {A_{11}}{A_{23}}} \\
          {{A_{21}}{A_{32}} - {A_{22}}{A_{31}}}&{{A_{12}}{A_{31}} - {A_{11}}{A_{32}}}&{{A_{11}}{A_{22}} - {A_{12}}{A_{21}}}
        \end{array}} \right]
    """

    qp_jacobi_invs = []
    for A, det_A in zip(qp_jacobis, qp_jacobi_dets):
        if A.shape == (2, 2):
            qp_jacobi_invs.append(array([[A[1][1], -A[0][1]], [-A[1][0], A[0][0]]]) / det_A)
        elif A.shape == (3, 3):
            qp_jacobi_invs.append(array([[(A[1][1] * A[2][2] - A[1][2] * A[2][1]),
                                          (A[0][2] * A[2][1] - A[0][1] * A[2][2]),
                                          (A[0][1] * A[1][2] - A[0][2] * A[1][1])],
                                         [(A[1][2] * A[2][0] - A[1][0] * A[2][2]),
                                          (A[0][0] * A[2][2] - A[0][2] * A[2][0]),
                                          (A[0][2] * A[1][0] - A[0][0] * A[1][2])],
                                         [(A[1][0] * A[2][1] - A[1][1] * A[2][0]),
                                          (A[0][1] * A[2][0] - A[0][0] * A[2][1]),
                                          (A[0][0] * A[1][1] - A[0][1] * A[1][0])]]) / det_A)
        else:
            return inv(qp_jacobis)
    return array(qp_jacobi_invs)


def get_transformation(u: ndarray, v: ndarray, w: ndarray,
                       u_prime: ndarray, v_prime: ndarray, w_prime: ndarray) -> ndarray:
    r"""
    **计算空间变换矩阵**

    :param u: :math:`\left( {{{{\mathbf{\hat e}}}_1},{{{\mathbf{\hat e}}}_2},{{{\mathbf{\hat e}}}_3}} \right)` 坐标系下的1号矢量
    :type u: ndarray

    :param v: :math:`\left( {{{{\mathbf{\hat e}}}_1},{{{\mathbf{\hat e}}}_2},{{{\mathbf{\hat e}}}_3}} \right)` 坐标系下的2号矢量
    :type v: ndarray

    :param w: :math:`\left( {{{{\mathbf{\hat e}}}_1},{{{\mathbf{\hat e}}}_2},{{{\mathbf{\hat e}}}_3}} \right)` 坐标系下的3号矢量
    :type w: ndarray

    :param u_prime: :math:`\left( {{{{\mathbf{\hat e'}}}_1},{{{\mathbf{\hat e'}}}_2},{{{\mathbf{\hat e'}}}_3}} \right)` 坐标系下的1号矢量
    :type u_prime: ndarray

    :param v_prime: :math:`\left( {{{{\mathbf{\hat e'}}}_1},{{{\mathbf{\hat e'}}}_2},{{{\mathbf{\hat e'}}}_3}} \right)` 坐标系下的2号矢量
    :type v_prime: ndarray

    :param w_prime: :math:`\left( {{{{\mathbf{\hat e'}}}_1},{{{\mathbf{\hat e'}}}_2},{{{\mathbf{\hat e'}}}_3}} \right)` 坐标系下的3号矢量
    :type w_prime: ndarray

    :return: 空间变换矩阵（线性）
    :rtype: ndarray

    设 :math:`\left( {{{{\mathbf{\hat e}}}_1},{{{\mathbf{\hat e}}}_2},{{{\mathbf{\hat e}}}_3}} \right)` 和 :math:`\left( {{{{\mathbf{\hat e'}}}_1},{{{\mathbf{\hat e'}}}_2},{{{\mathbf{\hat e'}}}_3}} \right)` 是空间 :math:`{{\mathbb{R}}^3}` 的两组基。
    如果矩阵 :math:`\mathbf{T}` 描述了两组基对应的线性变换 :math:`{{\mathbb{R}}^3} \Rightarrow {{\mathbb{R}}^3}`，则对于空间中的三个任意矢量有：

    .. math::
        \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right]\left\{ {\begin{array}{*{20}{c}}
          {{u_1}} \\
          {{u_2}} \\
          {{u_3}}
        \end{array}} \right\} = \left\{ {\begin{array}{*{20}{c}}
          {{{u'}_1}} \\
          {{{u'}_2}} \\
          {{{u'}_3}}
        \end{array}} \right\}

    .. math::
        \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right]\left\{ {\begin{array}{*{20}{c}}
          {{v_1}} \\
          {{v_2}} \\
          {{v_3}}
        \end{array}} \right\} = \left\{ {\begin{array}{*{20}{c}}
          {{{v'}_1}} \\
          {{{v'}_2}} \\
          {{{v'}_3}}
        \end{array}} \right\}

    .. math::
        \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right]\left\{ {\begin{array}{*{20}{c}}
          {{w_1}} \\
          {{w_2}} \\
          {{w_3}}
        \end{array}} \right\} = \left\{ {\begin{array}{*{20}{c}}
          {{{w'}_1}} \\
          {{{w'}_2}} \\
          {{{w'}_3}}
        \end{array}} \right\}

    合并可得：

    .. math::
        \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right]\left[ {\begin{array}{*{20}{c}}
          {{u_1}}&{{v_1}}&{{w_1}} \\
          {{u_2}}&{{v_2}}&{{w_2}} \\
          {{u_3}}&{{v_3}}&{{w_3}}
        \end{array}} \right] = \left[ {\begin{array}{*{20}{c}}
          {{{u'}_1}}&{{{v'}_1}}&{{{w'}_1}} \\
          {{{u'}_2}}&{{{v'}_2}}&{{{w'}_2}} \\
          {{{u'}_3}}&{{{v'}_3}}&{{{w'}_3}}
        \end{array}} \right]

    因此变换矩阵 :math:`\mathbf{T}` 可以表示为：

    .. math::
        \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right] = \left[ {\begin{array}{*{20}{c}}
          {{{u'}_1}}&{{{v'}_1}}&{{{w'}_1}} \\
          {{{u'}_2}}&{{{v'}_2}}&{{{w'}_2}} \\
          {{{u'}_3}}&{{{v'}_3}}&{{{w'}_3}}
        \end{array}} \right]{\left[ {\begin{array}{*{20}{c}}
          {{u_1}}&{{v_1}}&{{w_1}} \\
          {{u_2}}&{{v_2}}&{{w_2}} \\
          {{u_3}}&{{v_3}}&{{w_3}}
        \end{array}} \right]^{ - 1}}

    记为：

    .. math::
        {\mathbf{T}} = {\mathbf{A'}} \cdot {\left( {\mathbf{A}} \right)^{ - 1}}

    特殊的，对于两个空间直角坐标系的映射，即 :math:`\left( {{{{\mathbf{\hat e}}}_1},{{{\mathbf{\hat e}}}_2},{{{\mathbf{\hat e}}}_3}} \right)` 和 :math:`\left( {{{{\mathbf{\hat e'}}}_1},{{{\mathbf{\hat e'}}}_2},{{{\mathbf{\hat e'}}}_3}} \right)` 均为单位正交基，且变换需要满足  :math:`{{\mathbb{R}}^3} \Rightarrow {{\mathbb{R}}^3}` ，则 :math:`{\mathbf{A}}` 和  :math:`{{\mathbf{A'}}}` 必须满秩且为正交矩阵。
    """

    A = array([u, v, w])
    A_prime = array([u_prime, v_prime, w_prime])
    T = dot(A_prime, inv(A))
    return T


def voigt_array_to_tensor(voigt_array: ndarray, dimension: int) -> ndarray:
    r"""
    **Voigt记法数组转换为2阶张量**

    :param voigt_array: Voigt记法数组
    :type voigt_array: ndarray

    :param dimension: 空间维度
    :type dimension: int

    :return: 2阶张量
    :rtype: ndarray

    映射方式：

    .. math::
        \left\{ {\begin{array}{*{20}{c}}
          {{T_{11}}} \\
          {{T_{22}}} \\
          {{T_{12}}}
        \end{array}} \right\} \to \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}} \\
          {{T_{21}}}&{{T_{22}}}
        \end{array}} \right]

    .. math::
        \left\{ {\begin{array}{*{20}{c}}
          {{T_{11}}} \\
          {{T_{22}}} \\
          {{T_{33}}} \\
          {{T_{12}}} \\
          {{T_{13}}} \\
          {{T_{23}}}
        \end{array}} \right\} \to \left[ {\begin{array}{*{20}{c}}
          {{T_{11}}}&{{T_{12}}}&{{T_{13}}} \\
          {{T_{21}}}&{{T_{22}}}&{{T_{23}}} \\
          {{T_{31}}}&{{T_{32}}}&{{T_{33}}}
        \end{array}} \right]
    """

    tensor = zeros(shape=(dimension, dimension))
    if dimension == 2:
        tensor[0, 0] = voigt_array[0]
        tensor[1, 1] = voigt_array[1]
        tensor[0, 1] = voigt_array[2]
        tensor[1, 0] = voigt_array[2]
    elif dimension == 3:
        tensor[0, 0] = voigt_array[0]
        tensor[1, 1] = voigt_array[1]
        tensor[2, 2] = voigt_array[2]
        tensor[0, 1] = voigt_array[3]
        tensor[0, 2] = voigt_array[4]
        tensor[1, 2] = voigt_array[5]
        tensor[1, 0] = voigt_array[3]
        tensor[2, 0] = voigt_array[4]
        tensor[2, 1] = voigt_array[5]
    else:
        raise NotImplementedError(error_style(f'unsupported dimension {dimension}'))
    return tensor


def get_voigt_transformation(transformation: ndarray) -> ndarray:
    """
    **获取旋转矩阵的Voigt形式**

    :param transformation: Voigt记法数组
    :type transformation: ndarray

    :return: voigt_transformation
    :rtype: ndarray
    """

    voigt_transformation = zeros(shape=(6, 6), dtype=DTYPE)

    a11 = transformation[0][0]
    a12 = transformation[0][1]
    a13 = transformation[0][2]
    a21 = transformation[1][0]
    a22 = transformation[1][1]
    a23 = transformation[1][2]
    a31 = transformation[2][0]
    a32 = transformation[2][1]
    a33 = transformation[2][2]

    voigt_transformation[0][0] = a11 ** 2
    voigt_transformation[0][1] = a12 ** 2
    voigt_transformation[0][2] = a13 ** 2
    voigt_transformation[0][3] = 2 * a11 * a12
    voigt_transformation[0][4] = 2 * a11 * a13
    voigt_transformation[0][5] = 2 * a12 * a13

    voigt_transformation[1][0] = a21 ** 2
    voigt_transformation[1][1] = a22 ** 2
    voigt_transformation[1][2] = a23 ** 2
    voigt_transformation[1][3] = 2 * a21 * a22
    voigt_transformation[1][4] = 2 * a21 * a23
    voigt_transformation[1][5] = 2 * a22 * a23

    voigt_transformation[2][0] = a31 ** 2
    voigt_transformation[2][1] = a32 ** 2
    voigt_transformation[2][2] = a33 ** 2
    voigt_transformation[2][3] = 2 * a31 * a32
    voigt_transformation[2][4] = 2 * a31 * a33
    voigt_transformation[2][5] = 2 * a32 * a33

    voigt_transformation[3][0] = a11 * a21
    voigt_transformation[3][1] = a12 * a22
    voigt_transformation[3][2] = a13 * a23
    voigt_transformation[3][3] = a12 * a21 + a11 * a22
    voigt_transformation[3][4] = a13 * a21 + a11 * a23
    voigt_transformation[3][5] = a13 * a22 + a12 * a23

    voigt_transformation[4][0] = a11 * a31
    voigt_transformation[4][1] = a12 * a32
    voigt_transformation[4][2] = a13 * a33
    voigt_transformation[4][3] = a12 * a31 + a11 * a32
    voigt_transformation[4][4] = a13 * a31 + a11 * a33
    voigt_transformation[4][5] = a13 * a32 + a12 * a33

    voigt_transformation[5][0] = a21 * a31
    voigt_transformation[5][1] = a22 * a32
    voigt_transformation[5][2] = a23 * a33
    voigt_transformation[5][3] = a22 * a31 + a21 * a32
    voigt_transformation[5][4] = a23 * a31 + a21 * a33
    voigt_transformation[5][5] = a23 * a32 + a22 * a33

    return voigt_transformation


# def get_decompose_energy(strain: ndarray, solid_material_data0: float, solid_material_data1: float, dimension: int):
def get_decompose_energy(strain: ndarray, stress: ndarray, dimension: int):
    r"""
    **获取正的应变能密度**

    定义一种基于应变能谱分解的应变能分解方式（miehe）分解方法

    :param strain: Voigt记法数组
    :type strain: ndarray

    :param stress: Voigt记法数组
    :type stress: ndarray

    :param dimension: 单元维度
    :type dimension: int

    :return: energy_positive
    :rtype: ndarray

    :return: energy_negative
    :rtype: ndarray

    对于相场演化方程，如何选取合适的驱动力是尤为重要的。下面介绍一种通过能量分解的方法，在应变能中发掘驱动材料损伤的分量。

    2010年，miehe等[1]给出了一种基于应变张量谱分解的应变能分解方式，使用这种分解方式，可以得到一种拉压破坏模式不一致的相场模型，并且能够防止裂纹面的相互侵入问题。

    应变张量的谱分解表示为：

    .. math::
        {{\mathbf{\varepsilon}} }{\text{ = }}\sum\limits_{i = 1}^3 { \varepsilon _i } {n_i } \otimes {n_i }

    其中， :math:`{\varepsilon _i}(i = 1,2,3)` 为三个主应变分量。一个变形张量可以分解为三个不包含剪切变形，只拉伸或压缩变形的状态，这三个方向称为主应变方向，对应的大小为主应变。
    主应变及其方向可以通过求解应变矩阵的特征根来获得。下面求解应变张量不变量与主应变分量[2]。定义一个二阶应变张量为：

    .. math::
        {\mathbf{\varepsilon }} = \left[ {\begin{array}{*{20}{c}}
          {{\varepsilon _{11}}}&{{\varepsilon _{12}}}&{{\varepsilon _{13}}} \\
          {{\varepsilon _{21}}}&{{\varepsilon _{22}}}&{{\varepsilon _{23}}} \\
          {{\varepsilon _{31}}}&{{\varepsilon _{32}}}&{{\varepsilon _{33}}}
        \end{array}} \right]

    令主应变为 :math:`{\varepsilon _e}(e = 1,2,3)` ，则 :math:`{\varepsilon _e}` 可由下列方程求得

    .. math::
        \left| {\begin{array}{*{20}{c}}
          {{\varepsilon _{11}} - {\varepsilon _e}}&{{\varepsilon _{12}}}&{{\varepsilon _{13}}} \\
          {{\varepsilon _{21}}}&{{\varepsilon _{22}} - {\varepsilon _e}}&{{\varepsilon _{23}}} \\
          {{\varepsilon _{31}}}&{{\varepsilon _{32}}}&{{\varepsilon _{33}} - {\varepsilon _e}}
        \end{array}} \right| = 0

    上式可写为如下形式：

    .. math::
        \varepsilon _e^3 - {J_1}\varepsilon _e^2 + {J_2}{\varepsilon _e} - {J_3} = 0

    其中，方程的三个根即为三个主应变 :math:`{\varepsilon _1}，{\varepsilon _2}，{\varepsilon _3}` 。通常按照 :math:`{\varepsilon _1} \geqslant {\varepsilon _2} \geqslant {\varepsilon _3}` 来排列。
    系数项 :math:`{J_1}, {J_2}, {J_3}` 分别称为应变张量的第一、第二、第三不变量。

    其中，第一不变量 :math:`{J_1}` 定义为应变张量的迹 :math:`{J_1} = trace({\mathbf{\varepsilon }})` ;

    .. math::
        {J_1} = {\varepsilon _{11}} + {\varepsilon _{22}} + {\varepsilon _{33}} = {\varepsilon _1} +
        {\varepsilon _2} + {\varepsilon _3} = trace({\mathbf{\varepsilon }})

    第二不变量 :math:`{J_2}` 通常称为等效应变， :math:`{J_2} = \frac{1}{2}\left[ {{{\left( {trace\left( {\mathbf{\varepsilon }} \right)} \right)}^2} - trace\left( {{{\mathbf{\varepsilon }}^2}} \right)} \right]` ;

    .. math::
        {J_2} = \left| {\begin{array}{*{20}{c}}
          {{\varepsilon _{22}}}&{{\varepsilon _{23}}} \\
          {{\varepsilon _{32}}}&{{\varepsilon _{33}}}
        \end{array}} \right| + \left| {\begin{array}{*{20}{c}}
          {{\varepsilon _{11}}}&{{\varepsilon _{13}}} \\
          {{\varepsilon _{31}}}&{{\varepsilon _{33}}}
        \end{array}} \right| + \left| {\begin{array}{*{20}{c}}
          {{\varepsilon _{11}}}&{{\varepsilon _{12}}} \\
          {{\varepsilon _{21}}}&{{\varepsilon _{22}}}
        \end{array}} \right| =
        {\varepsilon _1}{\varepsilon _2} + {\varepsilon _2}{\varepsilon _3} + {\varepsilon _3}{\varepsilon _1} =
        \frac{1}{2}\left[ {{{\left( {trace\left( {\mathbf{\varepsilon }} \right)} \right)}^2} - trace\left( {{{\mathbf{\varepsilon }}^2}} \right)} \right]

    第三不变量 :math:`{J_3}` 定义为矩阵行列式， :math:`{J_3} = \det ({\mathbf{\varepsilon }})` 。

    .. math::
        {J_3} = \left| {\begin{array}{*{20}{c}}
          {{\varepsilon _{11}}}&{{\varepsilon _{12}}}&{{\varepsilon _{13}}} \\
          {{\varepsilon _{21}}}&{{\varepsilon _{22}}}&{{\varepsilon _{23}}} \\
          {{\varepsilon _{31}}}&{{\varepsilon _{32}}}&{{\varepsilon _{33}}}
        \end{array}} \right| = {\varepsilon _1}{\varepsilon _2}{\varepsilon _3} = \det ({\mathbf{\varepsilon }})

    用主应变分量可将应变能密度函数表示为：

    .. math::
        \psi ({\mathbf{\varepsilon }}) = \frac{1}{2}\lambda {\left( {{\varepsilon _1} + {\varepsilon _2} +
        {\varepsilon _3}} \right)^2} + \mu \left( {\varepsilon _1^2 + \varepsilon _2^2 + \varepsilon _3^2} \right)

    将上面的名义应变能密度函数分解为受拉和受压两个部分：

    .. math::
        \psi  = {\psi ^ + } + {\psi ^ - }

    对于各向同性线弹性材料，应变能密度函数可用应变张量表示为：

    .. math::
        {\psi ^ \pm }({\mathbf{\varepsilon }}) = \frac{1}{2}\lambda \left\langle {trace\left( {\mathbf{\varepsilon }} \right)} \right\rangle _ \pm ^2 +
        \mu {\text{ }}trace\left( {\left\langle {\varepsilon } \right\rangle _ \pm ^2} \right)

    即：

    .. math::
        {\psi ^ \pm }({\mathbf{\varepsilon }}) = \frac{1}{2}\lambda \left\langle {{\varepsilon _1} + {\varepsilon _2} + {\varepsilon _3}} \right\rangle _ \pm ^2 +
        \mu \left( {\left\langle {{\varepsilon _1}} \right\rangle _ \pm ^2 + \left\langle {{\varepsilon _2}} \right\rangle _ \pm ^2 + \left\langle {{\varepsilon _3}} \right\rangle _ \pm ^2} \right)

    其中， :math:`{\left\langle  \cdot  \right\rangle _ \pm }` 为Macaulay括号，表示为：

    .. math::
        {\left\langle x \right\rangle _ + }: = \left( {\left| x \right| + x} \right)/2;
        {\left\langle x \right\rangle _ - }: = \left( {\left| x \right| - x} \right)/2

    图像描述为::

                    <x>_{+}                                                <x>_{-}
                       |                                                      |
                       |                *              *                      |
                       |              *                  *                    |
                       |            *                       *                 |
                       |          *                            *              |
                       |        *                                 *           |
                       |      *                                      *        |
                       |    *                                           *     |
                       |  *                                                *  |
        ----***********0--x-------------------->   ---------------------------0**x**************-------->

    在分解后的应变能密度中，假设只有受拉应变能  :math:`{\psi ^ + }`  驱动相场的演化。而受压部分并不引起相场的演化，于是相场的驱动方程

    .. math::
        {F_d} =  - {w^{'}}(d) {\psi }( \varepsilon )

    可以改写为：

    .. math::
        {F_d} =  - {w^{'}}(d) {\psi ^ + }( \varepsilon )

    参考文献：

    [1] miehe, et.al., Thermodynamically consistent phase-field models of fracture: Variational principles and multi-field FE implementations. 2010.
    https://onlinelibrary.wiley.com/doi/abs/10.1002/nme.2861

    [2] 固体力学 [尹祥础 编] 2011年版

    """
    E = 1.0e5
    nu = 0.25
    mu = E / (2.0 * (1.0 + nu))
    lame = (E * nu) / ((1.0 + nu) * (1.0 - 2.0 * nu))

    strain = voigt_array_to_tensor(strain, dimension)

    # 得到主应变张量分量与特征方向
    principle_strain_value, principle_strain_vector = eig(strain)
    # print('principle_strain_value', principle_strain_value)

    # 1. 用向量乘法，计算高效，语法简洁
    # strain_positive = zeros(shape=(dimension, dimension))
    # strain_negative = zeros(shape=(dimension, dimension))
    #
    # for i in range(dimension):
    #     strain_positive += 0.5 * (principle_strain_value[i] + abs(principle_strain_value[i])) * \
    #                        tensordot(principle_strain_vector[:, i], principle_strain_vector[:, i], 0)
    #     strain_negative += 0.5 * (principle_strain_value[i] - abs(principle_strain_value[i])) * \
    #                        tensordot(principle_strain_vector[:, i], principle_strain_vector[:, i], 0)
    #     # tensordot(a,b,axes=1); 表示取a的最后几个维度,与b的前面几个维度相乘,再累加求和
    #     # 如： axes=1是指取a矩阵的最后一个维度，与b矩阵的第一个维度相乘
    #
    # energy_positive = 0.5 * lame * (0.5 * (strain.trace() + abs(strain.trace()))) ** 2 + \
    #                   mu * (strain_positive * strain_positive).trace()
    #
    # energy_negative = 0.5 * lame * (0.5 * (abs(strain.trace()) - strain.trace())) ** 2 + \
    #                   mu * (strain_negative * strain_negative).trace()
    #
    # print('energy_positive', energy_positive)
    # print('energy_negative', energy_negative)

    # energy = 0.5 * lame * (0.5 * strain.trace()) ** 2 + mu * (strain * strain).trace()

    # 2. 用函数定义，便于理解
    # 得到 <x>_{+} 与 <x>_{-}
    energy_positive = 0.0
    energy_negative = 0.0
    if dimension == 2:
        ep1 = max(principle_strain_value)
        ep2 = min(principle_strain_value)
        eq_sum = sum(principle_strain_value)
        if ep1 > 0.0:
            ep1_p = (abs(ep1) + ep1) / 2.0
            ep1_n = 0.0
        else:
            ep1_p = 0.0
            ep1_n = (abs(ep1) - ep1) / 2.0
        if ep2 > 0.0:
            ep2_p = (abs(ep2) + ep2) / 2.0
            ep2_n = 0.0
        else:
            ep2_p = 0.0
            ep2_n = (abs(ep2) - ep2) / 2.0
        if eq_sum > 0.0:
            eq_sum_p = (abs(eq_sum) + eq_sum) / 2.0
            eq_sum_n = 0.0
        else:
            eq_sum_p = 0.0
            eq_sum_n = (abs(eq_sum) - eq_sum) / 2.0
        energy_positive = lame / 2.0 * eq_sum_p ** 2.0 + mu * (ep1_p ** 2.0 + ep2_p ** 2.0)
        energy_negative = lame / 2.0 * eq_sum_n ** 2.0 + mu * (ep1_n ** 2.0 + ep2_n ** 2.0)

    elif dimension == 3:
        ep1 = max(principle_strain_value)
        ep3 = min(principle_strain_value)
        eq_sum = sum(principle_strain_value)
        ep2 = eq_sum - ep1 - ep3
        if ep1 > 0.0:
            ep1_p = (abs(ep1) + ep1) / 2.0
            ep1_n = 0.0
        else:
            ep1_p = 0.0
            ep1_n = (abs(ep1) - ep1) / 2.0
        if ep2 > 0.0:
            ep2_p = (abs(ep2) + ep2) / 2.0
            ep2_n = 0.0
        else:
            ep2_p = 0.0
            ep2_n = (abs(ep2) - ep2) / 2.0
        if ep3 > 0.0:
            ep3_p = (abs(ep3) + ep3) / 2.0
            ep3_n = 0.0
        else:
            ep3_p = 0.0
            ep3_n = (abs(ep3) - ep3) / 2.0
        if eq_sum > 0.0:
            eq_sum_p = (abs(eq_sum) + eq_sum) / 2.0
            eq_sum_n = 0.0
        else:
            eq_sum_p = 0.0
            eq_sum_n = (abs(eq_sum) - eq_sum) / 2.0
        energy_positive = lame / 2.0 * eq_sum_p ** 2.0 + mu * (ep1_p ** 2.0 + ep2_p ** 2.0 + ep3_p ** 2.0)
        energy_negative = lame / 2.0 * eq_sum_n ** 2.0 + mu * (ep1_n ** 2.0 + ep2_n ** 2.0 + ep3_n ** 2.0)

    # print('energy_positive', energy_positive)
    # print('energy_negative', energy_negative)

    return energy_positive, energy_negative


def operations_for_symtensor_antisymtensor(sym_tensor: ndarray, antisym_tensor: ndarray) -> ndarray:
    r"""
    **获取反对称张量乘以对称张量减去对称张量乘以反对称张量的结果**

    :param sym_tensor: :math:`\left[ {{a_{11}}{\text{ }}{a_{22}}{\text{ }}{a_{33}}{\text{ }}{a_{12}}{\text{ }}{a_{13}}{\text{ }}{a_{23}}} \right]` 采用 Vogit 记法的对称张量数组
    :type sym_tensor: ndarray

    :param antisym_tensor: :math:`\left[ {0{\text{ }}0{\text{ }}0{\text{ }}{b_{12}}{\text{ }}{b_{13}}{\text{ }}{b_{23}}} \right]` 采用类似 Vogit 记法的反对称张量数组
    :type antisym_tensor: ndarray

    :return: 反对称张量乘以对称张量减去对称张量乘以反对称张量的结果
    :rtype: ndarray

    首先分别将对称张量数组 :math:`\left[ {{a_{11}}{\text{ }}{a_{22}}{\text{ }}{a_{33}}{\text{ }}{a_{12}}{\text{ }}{a_{13}}{\text{ }}{a_{23}}} \right]` 和
    反对称张量数组 :math:`\left[ {0{\text{ }}0{\text{ }}0{\text{ }}{b_{12}}{\text{ }}{b_{13}}{\text{ }}{b_{23}}} \right]` 写成矩阵形式：

    .. math::
        A_{sym} = \left[ {\begin{array}{*{20}{c}}
          {{a_{11}}}&{{a_{12}}}&{{a_{13}}} \\
          {{a_{12}}}&{{a_{22}}}&{{a_{23}}} \\
          {{a_{13}}}&{{a_{23}}}&{{a_{33}}}
        \end{array}} \right],B_{antisym} = \left[ {\begin{array}{*{20}{c}}
          0&{{b_{12}}}&{{b_{13}}} \\
          { - {b_{12}}}&0&{{b_{23}}} \\
          { - {b_{13}}}&{ - {b_{23}}}&0
        \end{array}} \right]

    计算结果 :math:`{B_{antisym}}{A_{sym}} - {A_{sym}}{B_{antisym}}` 表示为：

    .. math::
        \begin{gathered}
          {B_{antisym}}{A_{sym}} - {A_{sym}}{B_{antisym}} = \left[ {\begin{array}{*{20}{c}}
          0&0&0&{2{b_{12}}}&{2{b_{13}}}&0 \\
          0&0&0&{ - 2{b_{12}}}&0&{2{b_{23}}} \\
          0&0&0&0&{ - 2{b_{13}}}&{ - 2{b_{23}}} \\
          { - {b_{12}}}&{{b_{12}}}&0&0&{{b_{23}}}&{{b_{13}}} \\
          { - {b_{13}}}&0&{{b_{13}}}&{ - {b_{23}}}&0&{{b_{12}}} \\
          0&{ - {b_{23}}}&{{b_{23}}}&{ - {b_{13}}}&{ - {b_{12}}}&0
        \end{array}} \right]\left[ \begin{gathered}
          {a_{11}} \hfill \\
          {a_{22}} \hfill \\
          {a_{33}} \hfill \\
          {a_{12}} \hfill \\
          {a_{13}} \hfill \\
          {a_{23}} \hfill \\
        \end{gathered}  \right] \hfill \\
           = \left[ \begin{gathered}
          2{b_{12}}{a_{12}} + 2{b_{13}}{a_{13}} \hfill \\
           - 2{b_{12}}{a_{12}} + 2{b_{23}}{a_{23}} \hfill \\
           - 2{b_{13}}{a_{13}} - 2{b_{23}}{a_{23}} \hfill \\
           - {b_{12}}{a_{11}} + {b_{12}}{a_{22}} + {b_{23}}{a_{13}} + {b_{13}}{a_{23}} \hfill \\
           - {b_{13}}{a_{11}} + {b_{13}}{a_{33}} - {b_{23}}{a_{12}} + {b_{12}}{a_{23}} \hfill \\
           - {b_{23}}{a_{22}} + {b_{23}}{a_{33}} - {b_{13}}{a_{12}} - {b_{12}}{a_{13}} \hfill \\
        \end{gathered}  \right] = T{A_{sym}} \hfill \\
        \end{gathered}
    """
    result = zeros(shape=(len(antisym_tensor), 6), dtype=DTYPE)

    for i in range(len(antisym_tensor)):
        T = array([[0, 0, 0, 2 * antisym_tensor[i, 3], 2 * antisym_tensor[i, 4], 0],
                   [0, 0, 0, -2 * antisym_tensor[i, 3], 0, 2 * antisym_tensor[i, 5]],
                   [0, 0, 0, 0, -2 * antisym_tensor[i, 4], -2 * antisym_tensor[i, 5]],
                   [-antisym_tensor[i, 3], antisym_tensor[i, 3], 0, 0, antisym_tensor[i, 5], antisym_tensor[i, 4]],
                   [-antisym_tensor[i, 4], 0, antisym_tensor[i, 4], -antisym_tensor[i, 5], 0, antisym_tensor[i, 3]],
                   [0, -antisym_tensor[i, 5], antisym_tensor[i, 5], -antisym_tensor[i, 4], -antisym_tensor[i, 3], 0]])
        result[i, :] = dot(T, sym_tensor)

    return result


if __name__ == '__main__':
    # u = array([1, 0, 0])
    # v = array([0, 1, 0])
    # w = array([0, 0, 1])
    #
    # u_prime = array([1, 0, 0])
    # v_prime = array([0, 1, 0])
    # w_prime = array([0, 0, 0.6187])
    # T = get_transformation(u, v, w, u_prime, v_prime, w_prime)
    #
    # n = array([1, 0, 0])
    # n = dot(T, n)
    # n *= 1.0 / norm(n)
    #
    # print(n)
    strain = np.array([-0.7, 0.6, 0.12])
    stress = np.array([-500, 643, 120])
    a, b = get_decompose_energy(strain, stress, 2)
    # print(a, b)
